/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*             This file is part of the program and software framework       */
/*                  UG --- Ubquity Generator Framework                       */
/*                                                                           */
/*    Copyright (C) 2010-2015 Konrad-Zuse-Zentrum                            */
/*                            fuer Informationstechnik Berlin                */
/*                                                                           */
/*  UG is distributed under the terms of the ZIB Academic Licence.           */
/*                                                                           */
/*  You should have received a copy of the ZIB Academic License              */
/*  along with UG; see the file COPYING. If not email to scip@zib.de.        */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/**@file    scipParaDiffSubproblemMpi.cpp
 * @brief   ScipParaDiffSubproblem extension for MPI communication.
 * @author  Yuji Shinano
 *
 *
 *
 */

/*---+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----8----+----9----+----0----+----1----+----2*/


#include <mpi.h>
#include "scipParaCommMpi.h"
#include "scipParaDiffSubproblemMpi.h"

using namespace UG;
using namespace ParaSCIP;

/** create ScipDiffSubproblemDatatype1 */
/************************************************
 * Currently, Datatype1 is not necessary.       *
 * I create this code for the future extension. *
 ************************************************/
MPI::Datatype
ScipParaDiffSubproblemMpi::createDatatype1(
      )
{

   int nBlocks = 0;

   MPI::Datatype datatype;

   MPI::Aint startAddress = 0;
   MPI::Aint address = 0;

#ifndef UG_DEBUG_SOLUTION
   int blockLengths[6];
   MPI::Aint displacements[6];
   MPI::Datatype types[6];
#else
   int blockLengths[7];
   MPI::Aint displacements[7];
   MPI::Datatype types[7];
#endif

   MPI_CALL_WITH_RET_VAL(
         startAddress,  MPI::Get_address( &localInfoIncluded )
   );
   blockLengths[nBlocks] = 1;
   displacements[nBlocks] = 0;
   types[nBlocks] = MPI::INT;
   nBlocks++;

   MPI_CALL_WITH_RET_VAL(
         address,  MPI::Get_address( &nBoundChanges )
   );
   blockLengths[nBlocks] = 1;
   displacements[nBlocks] = address - startAddress;
   types[nBlocks] = MPI::INT;
   nBlocks++;

   if( linearConss )
   {
      nLinearConss = linearConss->nLinearConss;
   }
   MPI_CALL_WITH_RET_VAL(
         address,  MPI::Get_address( &nLinearConss )
   );
   blockLengths[nBlocks] = 1;
   displacements[nBlocks] = address - startAddress;
   types[nBlocks] = MPI::INT;
   nBlocks++;

   if( boundDisjunctions )
   {
      nBoundDisjunctions = boundDisjunctions->nBoundDisjunctions;
   }
   MPI_CALL_WITH_RET_VAL(
         address,  MPI::Get_address( &nBoundDisjunctions )
   );
   blockLengths[nBlocks] = 1;
   displacements[nBlocks] = address - startAddress;
   types[nBlocks] = MPI::INT;
   nBlocks++;

   if( varBranchStats )
   {
      nVarBranchStats = varBranchStats->nVarBranchStats;
   }
   MPI_CALL_WITH_RET_VAL(
         address,  MPI::Get_address( &nVarBranchStats )
   );
   blockLengths[nBlocks] = 1;
   displacements[nBlocks] = address - startAddress;
   types[nBlocks] = MPI::INT;
   nBlocks++;

   if( varValues )
   {
      nVarValueVars = varValues->nVarValueVars;
   }
   MPI_CALL_WITH_RET_VAL(
         address,  MPI::Get_address( &nVarValueVars )
   );
   blockLengths[nBlocks] = 1;
   displacements[nBlocks] = address - startAddress;
   types[nBlocks] = MPI::INT;
   nBlocks++;

#ifdef UG_DEBUG_SOLUTION
   MPI_CALL_WITH_RET_VAL(
         address,  MPI::Get_address( &includeOptimalSol )
   );
   blockLengths[nBlocks] = 1;
   displacements[nBlocks] = address - startAddress;
   types[nBlocks] = MPI::INT;
   nBlocks++;
#endif

   MPI_CALL_WITH_RET_VAL(
         datatype, MPI::Datatype::Create_struct(nBlocks, blockLengths, displacements, types)
         );

   return datatype;
}

/** create ScipDiffSubproblemDatatype */
MPI::Datatype
ScipParaDiffSubproblemMpi::createDatatype2(
      bool memAllocNecessary
      )
{
   assert( nBoundChanges != 0 || nLinearConss != 0 || nBoundDisjunctions != 0 || nVarBranchStats != 0 || nVarValueVars != 0 );

   int nBlocks = 0;

   MPI::Datatype datatype;

   MPI::Aint startAddress = 0;
   MPI::Aint address = 0;

   int blockLengths[33];            // reserve maximum number of elements
   MPI::Aint displacements[33];     // reserve maximum number of elements
   MPI::Datatype types[33];         // reserve maximum number of elements

   if( nBoundChanges )
   {
      if( memAllocNecessary )
      {
         indicesAmongSolvers = new int[nBoundChanges];
         branchBounds = new SCIP_Real[nBoundChanges];
         boundTypes = new SCIP_BOUNDTYPE[nBoundChanges];
      }

      MPI_CALL_WITH_RET_VAL(
            startAddress,  MPI::Get_address( indicesAmongSolvers )
      );
      displacements[nBlocks] = 0;
      blockLengths[nBlocks] = nBoundChanges;
      types[nBlocks] = MPI::INT;
      nBlocks++;

      MPI_CALL_WITH_RET_VAL(
            address,  MPI::Get_address( branchBounds )
      );
      displacements[nBlocks] = address - startAddress;
      blockLengths[nBlocks] = nBoundChanges;
      types[nBlocks] = MPI::DOUBLE;
      nBlocks++;

      MPI_CALL_WITH_RET_VAL(
            address,  MPI::Get_address( boundTypes )
      );
      displacements[nBlocks] = address - startAddress;
      blockLengths[nBlocks] = nBoundChanges;
      types[nBlocks] = MPI::INT;
      nBlocks++;
   }

   if( nLinearConss )
   {
      if( memAllocNecessary )
      {
         linearConss = new ScipParaDiffSubproblemLinearCons();
         linearConss->nLinearConss = nLinearConss;
         linearConss->linearLhss = new SCIP_Real[nLinearConss];
         linearConss->linearRhss = new SCIP_Real[nLinearConss];
         linearConss->nLinearCoefs = new int[nLinearConss];
      }
      if( nBlocks == 0 )
      {
         MPI_CALL_WITH_RET_VAL(
               startAddress,  MPI::Get_address( linearConss->linearLhss )
         );
         displacements[nBlocks] = 0;
         blockLengths[nBlocks] = linearConss->nLinearConss;
         types[nBlocks] = MPI::DOUBLE;
         nBlocks++;
      }
      else
      {
         MPI_CALL_WITH_RET_VAL(
               address,  MPI::Get_address( linearConss->linearLhss )
         );
         displacements[nBlocks] = address - startAddress;;
         blockLengths[nBlocks] = linearConss->nLinearConss;
         types[nBlocks] = MPI::DOUBLE;
         nBlocks++;
      }

      MPI_CALL_WITH_RET_VAL(
            address,  MPI::Get_address( linearConss->linearRhss )
      );
      displacements[nBlocks] = address - startAddress;;
      blockLengths[nBlocks] = linearConss->nLinearConss;
      types[nBlocks] = MPI::DOUBLE;
      nBlocks++;

      MPI_CALL_WITH_RET_VAL(
            address,  MPI::Get_address( linearConss->nLinearCoefs )
      );
      displacements[nBlocks] = address - startAddress;;
      blockLengths[nBlocks] = linearConss->nLinearConss;
      types[nBlocks] = MPI::INT;
      nBlocks++;
   }

   if( nBoundDisjunctions )
   {
      if( memAllocNecessary )
      {
         boundDisjunctions = new ScipParaDiffSubproblemBoundDisjunctions();
         boundDisjunctions->nBoundDisjunctions = nBoundDisjunctions;
         boundDisjunctions->nVarsBoundDisjunction = new int[nBoundDisjunctions];
         boundDisjunctions->flagBoundDisjunctionInitial = new SCIP_Bool[nBoundDisjunctions];
         boundDisjunctions->flagBoundDisjunctionSeparate = new SCIP_Bool[nBoundDisjunctions];
         boundDisjunctions->flagBoundDisjunctionEnforce = new SCIP_Bool[nBoundDisjunctions];
         boundDisjunctions->flagBoundDisjunctionCheck = new SCIP_Bool[nBoundDisjunctions];
         boundDisjunctions->flagBoundDisjunctionPropagate = new SCIP_Bool[nBoundDisjunctions];
         boundDisjunctions->flagBoundDisjunctionLocal = new SCIP_Bool[nBoundDisjunctions];
         boundDisjunctions->flagBoundDisjunctionModifiable = new SCIP_Bool[nBoundDisjunctions];
         boundDisjunctions->flagBoundDisjunctionDynamic = new SCIP_Bool[nBoundDisjunctions];
         boundDisjunctions->flagBoundDisjunctionRemovable = new SCIP_Bool[nBoundDisjunctions];
         boundDisjunctions->flagBoundDisjunctionStickingatnode = new SCIP_Bool[nBoundDisjunctions];
      }
      if( nBlocks == 0 )
      {
         MPI_CALL_WITH_RET_VAL(
               startAddress,  MPI::Get_address( &(boundDisjunctions->nTotalVarsBoundDisjunctions) )
         );
         displacements[nBlocks] = 0;
         blockLengths[nBlocks] = 1;
         types[nBlocks] = MPI::INT;
         nBlocks++;
      }
      else
      {
         MPI_CALL_WITH_RET_VAL(
               address,  MPI::Get_address( &(boundDisjunctions->nTotalVarsBoundDisjunctions) )
         );
         displacements[nBlocks] = address - startAddress;;
         blockLengths[nBlocks] = 1;
         types[nBlocks] = MPI::INT;
         nBlocks++;
      }

      MPI_CALL_WITH_RET_VAL(
            address,  MPI::Get_address( boundDisjunctions->nVarsBoundDisjunction )
      );
      displacements[nBlocks] = address - startAddress;;
      blockLengths[nBlocks] = boundDisjunctions->nBoundDisjunctions;
      types[nBlocks] = MPI::UNSIGNED;
      nBlocks++;

      MPI_CALL_WITH_RET_VAL(
            address,  MPI::Get_address( boundDisjunctions->flagBoundDisjunctionInitial )
      );
      displacements[nBlocks] = address - startAddress;;
      blockLengths[nBlocks] = boundDisjunctions->nBoundDisjunctions;
      types[nBlocks] = MPI::UNSIGNED;
      nBlocks++;

      MPI_CALL_WITH_RET_VAL(
            address,  MPI::Get_address( boundDisjunctions->flagBoundDisjunctionSeparate )
      );
      displacements[nBlocks] = address - startAddress;;
      blockLengths[nBlocks] = boundDisjunctions->nBoundDisjunctions;
      types[nBlocks] = MPI::UNSIGNED;
      nBlocks++;

      MPI_CALL_WITH_RET_VAL(
            address,  MPI::Get_address( boundDisjunctions->flagBoundDisjunctionEnforce )
      );
      displacements[nBlocks] = address - startAddress;;
      blockLengths[nBlocks] = boundDisjunctions->nBoundDisjunctions;
      types[nBlocks] = MPI::UNSIGNED;
      nBlocks++;

      MPI_CALL_WITH_RET_VAL(
            address,  MPI::Get_address( boundDisjunctions->flagBoundDisjunctionCheck )
      );
      displacements[nBlocks] = address - startAddress;;
      blockLengths[nBlocks] = boundDisjunctions->nBoundDisjunctions;
      types[nBlocks] = MPI::UNSIGNED;
      nBlocks++;

      MPI_CALL_WITH_RET_VAL(
            address,  MPI::Get_address( boundDisjunctions->flagBoundDisjunctionPropagate )
      );
      displacements[nBlocks] = address - startAddress;;
      blockLengths[nBlocks] = boundDisjunctions->nBoundDisjunctions;
      types[nBlocks] = MPI::UNSIGNED;
      nBlocks++;

      MPI_CALL_WITH_RET_VAL(
            address,  MPI::Get_address( boundDisjunctions->flagBoundDisjunctionLocal )
      );
      displacements[nBlocks] = address - startAddress;;
      blockLengths[nBlocks] = boundDisjunctions->nBoundDisjunctions;
      types[nBlocks] = MPI::UNSIGNED;
      nBlocks++;

      MPI_CALL_WITH_RET_VAL(
            address,  MPI::Get_address( boundDisjunctions->flagBoundDisjunctionModifiable )
      );
      displacements[nBlocks] = address - startAddress;;
      blockLengths[nBlocks] = boundDisjunctions->nBoundDisjunctions;
      types[nBlocks] = MPI::UNSIGNED;
      nBlocks++;

      MPI_CALL_WITH_RET_VAL(
            address,  MPI::Get_address( boundDisjunctions->flagBoundDisjunctionDynamic )
      );
      displacements[nBlocks] = address - startAddress;;
      blockLengths[nBlocks] = boundDisjunctions->nBoundDisjunctions;
      types[nBlocks] = MPI::UNSIGNED;
      nBlocks++;

      MPI_CALL_WITH_RET_VAL(
            address,  MPI::Get_address( boundDisjunctions->flagBoundDisjunctionRemovable )
      );
      displacements[nBlocks] = address - startAddress;;
      blockLengths[nBlocks] = boundDisjunctions->nBoundDisjunctions;
      types[nBlocks] = MPI::UNSIGNED;
      nBlocks++;

      MPI_CALL_WITH_RET_VAL(
            address,  MPI::Get_address( boundDisjunctions->flagBoundDisjunctionStickingatnode )
      );
      displacements[nBlocks] = address - startAddress;;
      blockLengths[nBlocks] = boundDisjunctions->nBoundDisjunctions;
      types[nBlocks] = MPI::UNSIGNED;
      nBlocks++;

   }

   if( nVarBranchStats )
   {
      if( memAllocNecessary )
      {
         varBranchStats = new ScipParaDiffSubproblemVarBranchStats();
         varBranchStats->nVarBranchStats = nVarBranchStats;
         varBranchStats->idxBranchStatsVars = new int[nVarBranchStats];
         varBranchStats->downpscost = new SCIP_Real[nVarBranchStats];
         varBranchStats->uppscost = new SCIP_Real[nVarBranchStats];
         varBranchStats->downvsids = new SCIP_Real[nVarBranchStats];
         varBranchStats->upvsids = new SCIP_Real[nVarBranchStats];
         varBranchStats->downconflen = new SCIP_Real[nVarBranchStats];
         varBranchStats->upconflen = new SCIP_Real[nVarBranchStats];
         varBranchStats->downinfer = new SCIP_Real[nVarBranchStats];
         varBranchStats->upinfer = new SCIP_Real[nVarBranchStats];
         varBranchStats->downcutoff = new SCIP_Real[nVarBranchStats];
         varBranchStats->upcutoff = new SCIP_Real[nVarBranchStats];
      }

      if( nBlocks == 0 )
      {
         MPI_CALL_WITH_RET_VAL(
               startAddress,  MPI::Get_address( &(varBranchStats->offset) )
         );
         displacements[nBlocks] = 0;
         blockLengths[nBlocks] = 1;
         types[nBlocks] = MPI::INT;
         nBlocks++;
      }
      else
      {
         MPI_CALL_WITH_RET_VAL(
               address,  MPI::Get_address( &(varBranchStats->offset) )
         );
         displacements[nBlocks] = address - startAddress;;
         blockLengths[nBlocks] = 1;
         types[nBlocks] = MPI::INT;
         nBlocks++;
      }

      MPI_CALL_WITH_RET_VAL(
            address,  MPI::Get_address( varBranchStats->idxBranchStatsVars )
      );
      displacements[nBlocks] = address - startAddress;;
      blockLengths[nBlocks] = varBranchStats->nVarBranchStats;
      types[nBlocks] = MPI::INT;
      nBlocks++;

      MPI_CALL_WITH_RET_VAL(
            address,  MPI::Get_address( varBranchStats->downpscost )
      );
      displacements[nBlocks] = address - startAddress;;
      blockLengths[nBlocks] = varBranchStats->nVarBranchStats;
      types[nBlocks] = MPI::DOUBLE;
      nBlocks++;

      MPI_CALL_WITH_RET_VAL(
            address,  MPI::Get_address( varBranchStats->uppscost )
      );
      displacements[nBlocks] = address - startAddress;;
      blockLengths[nBlocks] = varBranchStats->nVarBranchStats;
      types[nBlocks] = MPI::DOUBLE;
      nBlocks++;

      MPI_CALL_WITH_RET_VAL(
            address,  MPI::Get_address( varBranchStats->downvsids )
      );
      displacements[nBlocks] = address - startAddress;;
      blockLengths[nBlocks] = varBranchStats->nVarBranchStats;
      types[nBlocks] = MPI::DOUBLE;
      nBlocks++;

      MPI_CALL_WITH_RET_VAL(
            address,  MPI::Get_address( varBranchStats->upvsids )
      );
      displacements[nBlocks] = address - startAddress;;
      blockLengths[nBlocks] = varBranchStats->nVarBranchStats;
      types[nBlocks] = MPI::DOUBLE;
      nBlocks++;

      MPI_CALL_WITH_RET_VAL(
            address,  MPI::Get_address( varBranchStats->downconflen )
      );
      displacements[nBlocks] = address - startAddress;;
      blockLengths[nBlocks] = varBranchStats->nVarBranchStats;
      types[nBlocks] = MPI::DOUBLE;
      nBlocks++;

      MPI_CALL_WITH_RET_VAL(
            address,  MPI::Get_address( varBranchStats->upconflen )
      );
      displacements[nBlocks] = address - startAddress;;
      blockLengths[nBlocks] = varBranchStats->nVarBranchStats;
      types[nBlocks] = MPI::DOUBLE;
      nBlocks++;

      MPI_CALL_WITH_RET_VAL(
            address,  MPI::Get_address( varBranchStats->downinfer )
      );
      displacements[nBlocks] = address - startAddress;;
      blockLengths[nBlocks] = varBranchStats->nVarBranchStats;
      types[nBlocks] = MPI::DOUBLE;
      nBlocks++;

      MPI_CALL_WITH_RET_VAL(
            address,  MPI::Get_address( varBranchStats->upinfer )
      );
      displacements[nBlocks] = address - startAddress;;
      blockLengths[nBlocks] = varBranchStats->nVarBranchStats;
      types[nBlocks] = MPI::DOUBLE;
      nBlocks++;

      MPI_CALL_WITH_RET_VAL(
            address,  MPI::Get_address( varBranchStats->downcutoff )
      );
      displacements[nBlocks] = address - startAddress;;
      blockLengths[nBlocks] = varBranchStats->nVarBranchStats;
      types[nBlocks] = MPI::DOUBLE;
      nBlocks++;

      MPI_CALL_WITH_RET_VAL(
            address,  MPI::Get_address( varBranchStats->upcutoff )
      );
      displacements[nBlocks] = address - startAddress;;
      blockLengths[nBlocks] = varBranchStats->nVarBranchStats;
      types[nBlocks] = MPI::DOUBLE;
      nBlocks++;

   }

   if( nVarValueVars )
   {
      if( memAllocNecessary )
      {
         varValues = new ScipParaDiffSubproblemVarValues();
         varValues->nVarValueVars = nVarValueVars;
         varValues->idxVarValueVars = new int[nVarValueVars];
         varValues->nVarValueValues = new int[nVarValueVars];
      }

      if( nBlocks == 0 )
      {
         MPI_CALL_WITH_RET_VAL(
               startAddress,  MPI::Get_address( &(varValues->nVarValues) )
         );
         displacements[nBlocks] = 0;
         blockLengths[nBlocks] = 1;
         types[nBlocks] = MPI::INT;
         nBlocks++;
      }
      else
      {
         MPI_CALL_WITH_RET_VAL(
               address,  MPI::Get_address( &(varValues->nVarValues) )
         );
         displacements[nBlocks] = address - startAddress;;
         blockLengths[nBlocks] = 1;
         types[nBlocks] = MPI::INT;
         nBlocks++;
      }

      MPI_CALL_WITH_RET_VAL(
            address,  MPI::Get_address( varValues->idxVarValueVars )
      );
      displacements[nBlocks] = address - startAddress;;
      blockLengths[nBlocks] = varValues->nVarValueVars;
      types[nBlocks] = MPI::INT;
      nBlocks++;

      MPI_CALL_WITH_RET_VAL(
            address,  MPI::Get_address( varValues->nVarValueValues )
      );
      displacements[nBlocks] = address - startAddress;;
      blockLengths[nBlocks] = varValues->nVarValueVars;
      types[nBlocks] = MPI::INT;
      nBlocks++;
   }

   assert(nBlocks <= 33);

   MPI_CALL_WITH_RET_VAL(
         datatype, MPI::Datatype::Create_struct(nBlocks, blockLengths, displacements, types)
   );

   return datatype;
}

/** create ScipDiffSubproblemDatatype */
MPI::Datatype
ScipParaDiffSubproblemMpi::createDatatype3(
      bool memAllocNecessary
      )
{
   assert( nLinearConss != 0 || nBoundDisjunctions != 0 || nVarValueVars != 0 );

   int nBlocks = 0;

   MPI::Datatype datatype;

   MPI::Aint startAddress = 0;
   MPI::Aint address = 0;

   int nVarValueBlock = 0;
   if( nVarValueVars > 0 )
   {
      for(int i = 0; i <  varValues->nVarValueVars; i++ )
      {
         if(  varValues->nVarValueValues[i] > 0 )
         {
            nVarValueBlock += 9;
         }
      }
   }

   int *blockLengths = new int[(nLinearConss*2) + (nBoundDisjunctions*3) + nVarValueBlock];
   MPI::Aint *displacements = new MPI::Aint[(nLinearConss*2) + (nBoundDisjunctions*3) + nVarValueBlock];
   MPI::Datatype *types = new MPI::Datatype[(nLinearConss*2) + (nBoundDisjunctions*3) + nVarValueBlock];

   if( memAllocNecessary && nLinearConss > 0 )
   {
      assert(linearConss);
      linearConss->linearCoefs = new SCIP_Real*[nLinearConss];
      linearConss->idxLinearCoefsVars = new int*[nLinearConss];
   }

   for(int i = 0; i < nLinearConss; i++ )
   {
      if( memAllocNecessary )
      {
         linearConss->linearCoefs[i] = new SCIP_Real[linearConss->nLinearCoefs[i]];
         linearConss->idxLinearCoefsVars[i] = new int[linearConss->nLinearCoefs[i]];
      }
      if( nBlocks == 0 )
      {
         MPI_CALL_WITH_RET_VAL(
               startAddress,  MPI::Get_address( linearConss->linearCoefs[i] )
         );
         displacements[nBlocks] = 0;
         blockLengths[nBlocks] = linearConss->nLinearCoefs[i];
         types[nBlocks] = MPI::DOUBLE;
         nBlocks++;
      }
      else
      {
         MPI_CALL_WITH_RET_VAL(
               address,  MPI::Get_address( linearConss->linearCoefs[i] )
         );
         displacements[nBlocks] = address - startAddress;
         blockLengths[nBlocks] = linearConss->nLinearCoefs[i];
         types[nBlocks] = MPI::DOUBLE;
         nBlocks++;
      }

      MPI_CALL_WITH_RET_VAL(
            address,  MPI::Get_address( linearConss->idxLinearCoefsVars[i] )
      );
      displacements[nBlocks] = address - startAddress;
      blockLengths[nBlocks] = linearConss->nLinearCoefs[i];
      types[nBlocks] = MPI::INT;
      nBlocks++;
   }

   if( memAllocNecessary && nBoundDisjunctions > 0 )
   {
      assert( boundDisjunctions );
      boundDisjunctions->idxBoundDisjunctionVars = new int*[nBoundDisjunctions];
      boundDisjunctions->boundTypesBoundDisjunction = new SCIP_BOUNDTYPE*[nBoundDisjunctions];
      boundDisjunctions->boundsBoundDisjunction = new SCIP_Real*[nBoundDisjunctions];
   }

   for( int i = 0; i < nBoundDisjunctions; i++ )
   {
      if( memAllocNecessary )
      {
         boundDisjunctions->idxBoundDisjunctionVars[i] = new int[boundDisjunctions->nVarsBoundDisjunction[i]];
         boundDisjunctions->boundTypesBoundDisjunction[i] = new SCIP_BOUNDTYPE[boundDisjunctions->nVarsBoundDisjunction[i]];
         boundDisjunctions->boundsBoundDisjunction[i] = new SCIP_Real[boundDisjunctions->nVarsBoundDisjunction[i]];
      }
      if( nBlocks == 0 )
      {
         MPI_CALL_WITH_RET_VAL(
               startAddress,  MPI::Get_address( boundDisjunctions->idxBoundDisjunctionVars[i] )
         );
         displacements[nBlocks] = 0;
         blockLengths[nBlocks] = boundDisjunctions->nVarsBoundDisjunction[i];
         types[nBlocks] = MPI::INT;
         nBlocks++;
      }
      else
      {
         MPI_CALL_WITH_RET_VAL(
               address,  MPI::Get_address( boundDisjunctions->idxBoundDisjunctionVars[i] )
         );
         displacements[nBlocks] = address - startAddress;
         blockLengths[nBlocks] = boundDisjunctions->nVarsBoundDisjunction[i];
         types[nBlocks] = MPI::INT;
         nBlocks++;
      }

      MPI_CALL_WITH_RET_VAL(
            address,  MPI::Get_address( boundDisjunctions->boundTypesBoundDisjunction[i] )
      );
      displacements[nBlocks] = address - startAddress;
      blockLengths[nBlocks] = boundDisjunctions->nVarsBoundDisjunction[i];
      types[nBlocks] = MPI::UNSIGNED;   // actual SCIP_BoundType is enum
      nBlocks++;

      MPI_CALL_WITH_RET_VAL(
            address,  MPI::Get_address( boundDisjunctions->boundsBoundDisjunction[i] )
      );
      displacements[nBlocks] = address - startAddress;
      blockLengths[nBlocks] = boundDisjunctions->nVarsBoundDisjunction[i];
      types[nBlocks] = MPI::DOUBLE;
      nBlocks++;

   }

   if( memAllocNecessary && nVarValueVars > 0 )
   {
      assert(varValues);
      varValues->varValue            = new SCIP_Real*[nVarValueVars];
      varValues->varValueDownvsids   = new SCIP_Real*[nVarValueVars];
      varValues->varVlaueUpvsids     = new SCIP_Real*[nVarValueVars];
      varValues->varValueDownconflen = new SCIP_Real*[nVarValueVars];
      varValues->varValueUpconflen   = new SCIP_Real*[nVarValueVars];
      varValues->varValueDowninfer   = new SCIP_Real*[nVarValueVars];
      varValues->varValueUpinfer     = new SCIP_Real*[nVarValueVars];
      varValues->varValueDowncutoff  = new SCIP_Real*[nVarValueVars];
      varValues->varValueUpcutoff    = new SCIP_Real*[nVarValueVars];
   }

   for(int i = 0; i < nVarValueVars; i++ )
   {
      if(varValues-> nVarValueValues[i] > 0 )
      {
         if( memAllocNecessary )
         {
            varValues->varValue[i]            = new SCIP_Real[varValues->nVarValueValues[i]];
            varValues->varValueDownvsids[i]   = new SCIP_Real[varValues->nVarValueValues[i]];
            varValues->varVlaueUpvsids[i]     = new SCIP_Real[varValues->nVarValueValues[i]];
            varValues->varValueDownconflen[i] = new SCIP_Real[varValues->nVarValueValues[i]];
            varValues->varValueUpconflen[i]   = new SCIP_Real[varValues->nVarValueValues[i]];
            varValues->varValueDowninfer[i]   = new SCIP_Real[varValues->nVarValueValues[i]];
            varValues->varValueUpinfer[i]     = new SCIP_Real[varValues->nVarValueValues[i]];
            varValues->varValueDowncutoff[i]  = new SCIP_Real[varValues->nVarValueValues[i]];
            varValues->varValueUpcutoff[i]    = new SCIP_Real[varValues->nVarValueValues[i]];
         }

         if( nBlocks == 0 )
         {
            MPI_CALL_WITH_RET_VAL(
                  startAddress,  MPI::Get_address( varValues->varValue[i] )
            );
            displacements[nBlocks] = 0;
            blockLengths[nBlocks] = varValues->nVarValueValues[i];
            types[nBlocks] = MPI::DOUBLE;
            nBlocks++;
         }
         else
         {
            MPI_CALL_WITH_RET_VAL(
                  address,  MPI::Get_address( varValues->varValue[i] )
            );
            displacements[nBlocks] = address - startAddress;
            blockLengths[nBlocks] = varValues->nVarValueValues[i];
            types[nBlocks] = MPI::DOUBLE;
            nBlocks++;
         }

         MPI_CALL_WITH_RET_VAL(
               address,  MPI::Get_address( varValues->varValueDownvsids[i] )
         );
         displacements[nBlocks] = address - startAddress;
         blockLengths[nBlocks] = varValues->nVarValueValues[i];
         types[nBlocks] = MPI::DOUBLE;
         nBlocks++;

         MPI_CALL_WITH_RET_VAL(
               address,  MPI::Get_address( varValues->varVlaueUpvsids[i] )
         );
         displacements[nBlocks] = address - startAddress;
         blockLengths[nBlocks] = varValues->nVarValueValues[i];
         types[nBlocks] = MPI::DOUBLE;
         nBlocks++;

         MPI_CALL_WITH_RET_VAL(
               address,  MPI::Get_address( varValues->varValueDownconflen[i] )
         );
         displacements[nBlocks] = address - startAddress;
         blockLengths[nBlocks] = varValues->nVarValueValues[i];
         types[nBlocks] = MPI::DOUBLE;
         nBlocks++;

         MPI_CALL_WITH_RET_VAL(
               address,  MPI::Get_address( varValues->varValueUpconflen[i] )
         );
         displacements[nBlocks] = address - startAddress;
         blockLengths[nBlocks] = varValues->nVarValueValues[i];
         types[nBlocks] = MPI::DOUBLE;
         nBlocks++;

         MPI_CALL_WITH_RET_VAL(
               address,  MPI::Get_address( varValues->varValueDowninfer[i] )
         );
         displacements[nBlocks] = address - startAddress;
         blockLengths[nBlocks] = varValues->nVarValueValues[i];
         types[nBlocks] = MPI::DOUBLE;
         nBlocks++;

         MPI_CALL_WITH_RET_VAL(
               address,  MPI::Get_address( varValues->varValueUpinfer[i] )
         );
         displacements[nBlocks] = address - startAddress;
         blockLengths[nBlocks] = varValues->nVarValueValues[i];
         types[nBlocks] = MPI::DOUBLE;
         nBlocks++;

         MPI_CALL_WITH_RET_VAL(
               address,  MPI::Get_address( varValues->varValueDowncutoff[i] )
         );
         displacements[nBlocks] = address - startAddress;
         blockLengths[nBlocks] = varValues->nVarValueValues[i];
         types[nBlocks] = MPI::DOUBLE;
         nBlocks++;

         MPI_CALL_WITH_RET_VAL(
               address,  MPI::Get_address( varValues->varValueUpcutoff[i] )
         );
         displacements[nBlocks] = address - startAddress;
         blockLengths[nBlocks] = varValues->nVarValueValues[i];
         types[nBlocks] = MPI::DOUBLE;
         nBlocks++;

      }

   }

   MPI_CALL_WITH_RET_VAL(
         datatype, MPI::Datatype::Create_struct(nBlocks, blockLengths, displacements, types)
   );

   delete [] blockLengths;
   delete [] displacements;
   delete [] types;

   return datatype;
}

int
ScipParaDiffSubproblemMpi::bcast(ParaComm *comm, int root)
{
   DEF_PARA_COMM( commMpi, comm);
   MPI::Datatype datatype1;
   datatype1 = createDatatype1();
   datatype1.Commit();
   PARA_COMM_CALL(
      commMpi->bcast(&localInfoIncluded, 1, datatype1, root)
   );
   datatype1.Free();

   if( nBoundChanges > 0 || nLinearConss > 0 || nBoundDisjunctions > 0 || nVarBranchStats > 0 || nVarValueVars > 0 ){
      MPI::Datatype datatype2;
      if( comm->getRank() == root )
      {
         datatype2 = createDatatype2(false);
      }
      else
      {
         datatype2 = createDatatype2(true);
      }
      datatype2.Commit();
      if( nBoundChanges > 0 )
      {
         PARA_COMM_CALL(
               commMpi->bcast(indicesAmongSolvers, 1, datatype2, root)
         );
      }
      else if( nLinearConss > 0 )
      {
         PARA_COMM_CALL(
               commMpi->bcast(linearConss->linearLhss, 1, datatype2, root)
         );
      }
      else if( nBoundDisjunctions > 0 )
      {
         PARA_COMM_CALL(
               commMpi->bcast(&(boundDisjunctions->nTotalVarsBoundDisjunctions), 1, datatype2, root)
         );
      }
      else if( nVarBranchStats > 0 )
      {
         PARA_COMM_CALL(
               commMpi->bcast(&(varBranchStats->offset), 1, datatype2, root)
         );
      }
      else
      {
         PARA_COMM_CALL(
               commMpi->bcast(&(varValues->nVarValues), 1, datatype2, root)
         );
      }
      datatype2.Free();

      if( nLinearConss > 0 || nBoundDisjunctions > 0 || nVarValueVars > 0 )
      {
         MPI::Datatype datatype3;
         if( comm->getRank() == root )
         {
            datatype3 = createDatatype3(false);
         }
         else
         {
            datatype3 = createDatatype3(true);
         }
         datatype3.Commit();

         if( nLinearConss > 0 )
         {
            PARA_COMM_CALL(
                  commMpi->bcast(linearConss->linearCoefs[0], 1, datatype3, root)
            );
         }
         else if( nBoundDisjunctions > 0 )
         {
            PARA_COMM_CALL(
                  commMpi->bcast(boundDisjunctions->idxBoundDisjunctionVars[0], 1, datatype3, root)
            );
         }
         else
         {
            PARA_COMM_CALL(
                  commMpi->bcast(varValues->varValue[0], 1, datatype3, root)
            );
         }

         datatype3.Free();
      }
   }
   return 0;
}

int
ScipParaDiffSubproblemMpi::send(ParaComm *comm, int dest)
{
   DEF_PARA_COMM( commMpi, comm);
   MPI::Datatype datatype1;
   datatype1 = createDatatype1();
   datatype1.Commit();
   PARA_COMM_CALL(
      commMpi->send(&localInfoIncluded, 1, datatype1, dest, TagDiffSubproblem)
   );
   datatype1.Free();

   if( nBoundChanges > 0 || nLinearConss > 0 || nBoundDisjunctions > 0 || nVarBranchStats > 0 || nVarValueVars > 0 ){
      MPI::Datatype datatype2;
      datatype2 = createDatatype2(false);
      datatype2.Commit();
      if( nBoundChanges > 0 )
      {
         PARA_COMM_CALL(
               commMpi->send(indicesAmongSolvers, 1, datatype2, dest, TagDiffSubproblem1)
         );
      }
      else if( nLinearConss > 0 )
      {
         PARA_COMM_CALL(
               commMpi->send(linearConss->linearLhss, 1, datatype2, dest, TagDiffSubproblem1)
         );
      }
      else if( nBoundDisjunctions > 0 )
      {
         PARA_COMM_CALL(
               commMpi->send(&(boundDisjunctions->nTotalVarsBoundDisjunctions), 1, datatype2, dest, TagDiffSubproblem1)
         );
      }
      else if( nVarBranchStats > 0 )
      {
         PARA_COMM_CALL(
               commMpi->send(&(varBranchStats->offset), 1, datatype2, dest, TagDiffSubproblem1)
         );
      }
      else
      {
         PARA_COMM_CALL(
               commMpi->send(&(varValues->nVarValues), 1, datatype2, dest, TagDiffSubproblem1)
         );
      }
      datatype2.Free();
      if( nLinearConss > 0 || nBoundDisjunctions > 0  || nVarValueVars > 0 )
      {
         MPI::Datatype datatype3;
         datatype3 = createDatatype3(false);
         datatype3.Commit();

         if( nLinearConss > 0 )
         {
            PARA_COMM_CALL(
                  commMpi->send(linearConss->linearCoefs[0], 1, datatype3, dest, TagDiffSubproblem2)
            );
         }
         else if( nBoundDisjunctions > 0 )
         {
            PARA_COMM_CALL(
                  commMpi->send(boundDisjunctions->idxBoundDisjunctionVars[0], 1, datatype3, dest, TagDiffSubproblem2)
            );
         }
         else
         {
            PARA_COMM_CALL(
                  commMpi->send(varValues->varValue[0], 1, datatype3, dest, TagDiffSubproblem2)
            );
         }

         datatype3.Free();
      }
   }
   return 0;
}

int
ScipParaDiffSubproblemMpi::receive(ParaComm *comm, int source)
{
   DEF_PARA_COMM( commMpi, comm);
   MPI::Datatype datatype1;
   datatype1 = createDatatype1();
   datatype1.Commit();
   PARA_COMM_CALL(
      commMpi->receive(&localInfoIncluded, 1, datatype1, source, TagDiffSubproblem)
   );
   datatype1.Free();

   if( nBoundChanges > 0 || nLinearConss > 0 || nBoundDisjunctions > 0 || nVarBranchStats > 0 || nVarValueVars > 0 )
   {
      MPI::Datatype datatype2;
      datatype2 = createDatatype2(true);
      datatype2.Commit();
      if( nBoundChanges > 0 )
      {
         PARA_COMM_CALL(
               commMpi->receive(indicesAmongSolvers, 1, datatype2, source, TagDiffSubproblem1)
         );
      }
      else if( nLinearConss > 0 )
      {
         PARA_COMM_CALL(
               commMpi->receive(linearConss->linearLhss, 1, datatype2, source, TagDiffSubproblem1)
         );
      }
      else if( nBoundDisjunctions > 0 )
      {
         PARA_COMM_CALL(
               commMpi->receive(&(boundDisjunctions->nTotalVarsBoundDisjunctions), 1, datatype2, source, TagDiffSubproblem1)
         );
      }
      else if( nVarBranchStats > 0 )
      {
         PARA_COMM_CALL(
               commMpi->receive(&(varBranchStats->offset), 1, datatype2, source, TagDiffSubproblem1)
         );
      }
      else
      {
         PARA_COMM_CALL(
               commMpi->receive(&(varValues->nVarValues), 1, datatype2, source, TagDiffSubproblem1)
         );
      }
      datatype2.Free();
      if( nLinearConss > 0 || nBoundDisjunctions > 0 || nVarValueVars > 0 )
      {
         MPI::Datatype datatype3;
         datatype3 = createDatatype3(true);
         datatype3.Commit();
         if( nLinearConss > 0 )
         {
            PARA_COMM_CALL(
                  commMpi->receive(linearConss->linearCoefs[0], 1, datatype3, source, TagDiffSubproblem2)
            );
         }
         else if( nBoundDisjunctions > 0)
         {
            PARA_COMM_CALL(
                  commMpi->receive(boundDisjunctions->idxBoundDisjunctionVars[0], 1, datatype3, source, TagDiffSubproblem2)
            );
         }
         else
         {
            PARA_COMM_CALL(
                  commMpi->receive(varValues->varValue[0], 1, datatype3, source, TagDiffSubproblem2)
            );
         }

         datatype3.Free();
      }
   }
   return 0;
}

/** create clone of this object */
ScipParaDiffSubproblemMpi *
ScipParaDiffSubproblemMpi::clone(ParaComm *comm)
{
   return( new ScipParaDiffSubproblemMpi(this) );

}
