/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*             This file is part of the program and software framework       */
/*                  UG --- Ubquity Generator Framework                       */
/*                                                                           */
/*    Copyright (C) 2010-2015 Konrad-Zuse-Zentrum                            */
/*                            fuer Informationstechnik Berlin                */
/*                                                                           */
/*  UG is distributed under the terms of the ZIB Academic Licence.           */
/*                                                                           */
/*  You should have received a copy of the ZIB Academic License              */
/*  along with UG; see the file COPYING. If not email to scip@zib.de.        */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/**@file    scipParaInstancePth.h
 * @brief   ScipParaInstance extension for Pthreads communication.
 * @author  Yuji Shinano
 *
 *
 *
 */

/*---+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----8----+----9----+----0----+----1----+----2*/


#ifndef __SCIP_PARA_INSTANCE_PTH_H__
#define __SCIP_PARA_INSTANCE_PTH_H__

#include <cassert>
#include <cstring>
#include "ug/paraInstance.h"
#include "ug/paraCommPth.h"
#include "scipUserPlugins.h"
#include "scip/scip.h"
#include "scip/cons_linear.h"
#include "scipDiffParamSet.h"
#include "scipParaSolution.h"

namespace ParaSCIP
{

class ScipParaInstance : public UG::ParaInstance
{
protected:
   SCIP *scip;     // this pointer should point to the scip environment of LoadCoordinator, in case that
                   // LC does not have two scip environment
   virtual const char *getFileName() = 0;
   int           nVars;
   int           *mapToOriginalIndecies;  /**< array of indices to map to original problem's probindices
                                               LC does not have this map in geneal, it is in the transformed prob. in scip */
   SCIP          *orgScip;                /**< if LC has the abolve MAP, varialbes need to be converted to this scip */
public:
   /** constructor */
   ScipParaInstance(
         ) : scip(0), nVars(0), mapToOriginalIndecies(0), orgScip(0)
   {
   }

   ScipParaInstance(
		 SCIP *inScip
         ) : scip(inScip), nVars(0), mapToOriginalIndecies(0), orgScip(0)
   {
   }
   /** destractor */
   virtual ~ScipParaInstance(
         )
   {
      if( mapToOriginalIndecies )
      {
         delete [] mapToOriginalIndecies;
      }
      if( orgScip )    // If this instance has orgScip pointer, scip has to be freed
      {
         SCIPfree(&scip);
      }
   }

   /** convert an internal value to external value */
   double convertToExternalValue(double internalValue)
   {
      if( orgScip )
      {
         return SCIPretransformObj(orgScip, SCIPretransformObj(scip, internalValue) );
      }
      else
      {
         return SCIPretransformObj(scip, internalValue);
      }
   }

   /** convert an external value to internal value */
   double convertToInternalValue(double externalValue)
   {
      if( orgScip )
      {
         return SCIPtransformObj(scip, SCIPtransformObj(orgScip, externalValue));
      }
      else
      {
         return SCIPtransformObj(scip, externalValue);
      }
   }

   /** get solution vaues for the original problem */
   void getSolValuesForOriginalProblem(ScipParaSolution *sol, SCIP_Real *vals)
   {
      assert( orgScip );
      SCIP_SOL*  newsol;
      SCIP_CALL_ABORT( SCIPcreateSol(scip, &newsol, 0) );
      SCIP_VAR **vars = SCIPgetVars(scip);
      for( int i = 0; i < sol->getNVars(); i++ )
      {
         SCIP_CALL_ABORT( SCIPsetSolVal(scip, newsol, vars[sol->indexAmongSolvers(i)], sol->getValues()[i]) );
      }
      SCIP_CALL_ABORT(SCIPgetSolVals(scip, newsol, SCIPgetNVars(scip), SCIPgetVars(scip), vals) );
      SCIP_Bool success;
      SCIP_CALL_ABORT( SCIPaddSolFree(scip, &newsol, &success) );
   }

   /** create presolved problem instance that is solved by ParaSCIP form scip environment in this object */
   void copyScipEnvironment(
         SCIP **scip
         );

   SCIP *getScip(
         )
   {
      return scip;
   }

   /** create presolved problem instance that is solved by ParaSCIP */
   void createProblem(
        SCIP *scip,
        int  method,               // transferring method
        bool noPreprocessingInLC,  // LC preprocesing settings
        bool usetRootNodeCuts,
        ScipDiffParamSet  *scipDiffParamSetRoot,
        char *settingsNameLC,       // LC preprocesing settings
        char *isolname
        );

   /** stringfy ParaCalculationState */
   const std::string toString(
        )
   {
      return std::string("Should be written from scip environment.");
   }

   // int getOrigProbIndex(int index){ return mapToOriginalIndecies[index]; }
   const char *getProbName()
   {
      if( orgScip )
      {
         return SCIPgetProbName(orgScip);
      }
      else
      {
         return SCIPgetProbName(scip);
      }
   }

   int getNVars(){ return nVars; }
   SCIP_Real getVarLb(int i)
   {
      SCIP_VAR **vars = SCIPgetVars(scip);
      return SCIPvarGetLbGlobal(vars[i]);
   }
   SCIP_Real getVarUb(int i)
   {
      SCIP_VAR **vars = SCIPgetVars(scip);
      return SCIPvarGetUbGlobal(vars[i]);
   }
   int getVarType(int i)
   {
      SCIP_VAR **vars = SCIPgetVars(scip);
      return SCIPvarGetType(vars[i]);
   }
   SCIP_Real getObjCoef(int i)
   {
      SCIP_VAR **vars = SCIPgetVars(scip);
      return SCIPvarGetObj(vars[i]);
   }

   const char *getVarName(int i)
   {
      SCIP_VAR **vars = SCIPgetVars(scip);
      return SCIPvarGetName(vars[i]);
   }

   int getNConss()
   {
      return SCIPgetNConss(scip);
   }

   SCIP_Real getLhsLinear(int i)
   {
      SCIP_CONS **conss = SCIPgetConss(scip);
      SCIP_CONSHDLR* conshdlr = SCIPconsGetHdlr(conss[i]);
      assert( conshdlr != NULL );
      if( strcmp(SCIPconshdlrGetName(conshdlr),"linear") != 0 )
      {
         THROW_LOGICAL_ERROR2("invalid constraint type exists; consname = ", SCIPconshdlrGetName(conshdlr));
      }
      return SCIPgetLhsLinear(scip, conss[i]);
   }

   SCIP_Real getRhsLinear(int i)
   {
      SCIP_CONS **conss = SCIPgetConss(scip);
      SCIP_CONSHDLR* conshdlr = SCIPconsGetHdlr(conss[i]);
      assert( conshdlr != NULL );
      if( strcmp(SCIPconshdlrGetName(conshdlr),"linear") != 0 )
      {
         THROW_LOGICAL_ERROR2("invalid constraint type exists; consname = ", SCIPconshdlrGetName(conshdlr));
      }
      return SCIPgetRhsLinear(scip, conss[i]);
   }

   int getNVarsLinear(int i)
   {
      SCIP_CONS **conss = SCIPgetConss(scip);
      SCIP_CONSHDLR* conshdlr = SCIPconsGetHdlr(conss[i]);
      assert( conshdlr != NULL );
      if( strcmp(SCIPconshdlrGetName(conshdlr),"linear") != 0 )
      {
         THROW_LOGICAL_ERROR2("invalid constraint type exists; consname = ", SCIPconshdlrGetName(conshdlr));
      }
      return SCIPgetNVarsLinear(scip, conss[i]);
   }

   int getIdxLinearCoefVar(int i, int j) // i-th constraint, j-th variable
   {
      SCIP_CONS **conss = SCIPgetConss(scip);
      SCIP_VAR **vars = SCIPgetVarsLinear(scip, conss[i]);
      return SCIPvarGetProbindex(vars[j]);
   }

   SCIP_Real *getLinearCoefs(int i) // i-th constraint, j-th variable
   {
      SCIP_CONS **conss = SCIPgetConss(scip);
      return SCIPgetValsLinear(scip, conss[i]);
   }

   const char *getConsName(int i)
   {
      SCIP_CONS **conss = SCIPgetConss(scip);
      return SCIPconsGetName(conss[i]);
   }

   /** set user plugins */
   void setUserPlugins(ScipUserPlugins *inUi) { /** maybe called, no need to do anything */ }

   /** include user plugins */
   void includeUserPlugins(SCIP *inScip){/** should not be called **/}

   virtual void setFileName( const char *file ) = 0;

  bool isOriginalIndeciesMap() { return (mapToOriginalIndecies != 0); }

  int getOrigProbIndex(int index)
  {
     assert(mapToOriginalIndecies);
     return mapToOriginalIndecies[index];
  }

  int *extractOrigProbIndexMap()
  {
     assert(mapToOriginalIndecies);
     int *extract = mapToOriginalIndecies;
     mapToOriginalIndecies = 0;
     return extract;
  }

  SCIP *getParaInstanceScip()
  {
     assert( orgScip );
     return scip;
  }

};

}

namespace ParaSCIP
{

/** ScipInstancePth */
class ScipParaInstancePth : public ScipParaInstance
{
   const char *getFileName()
   {
      std::cout << "This function should name be used for FiberSCIP!" << std::endl;
      abort();
   }
public:
   /** constructor */
   ScipParaInstancePth(
         )
   {
   }

   /** constructor : only called from ScipInitiator */
   ScipParaInstancePth(
         SCIP *inScip,
         int method
         );

   /** destractor */
   ~ScipParaInstancePth(
         )
   {
   }

   /** broadcasts instance to all solvers */
   int bcast(UG::ParaComm *comm, int rank, int method);

   void setFileName( const char *file )
   {
      std::cout << "This function should name be used for FiberSCIP!" << std::endl;
      abort();
   }

};

typedef ScipParaInstancePth *ScipParaInstancePthPtr;

}

#endif  // __SCIP_PARA_INSTANCE_PTH_H__

